using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;

// VAMMusicLoader v1.0
//
// Loads an assetbundle with a single music file to be able to distribute licensed music

namespace VAMMusicLoader
{
    public class VAMMusicLoaderPlugin : MVRScript
    {	
		// Static string
		private static string PLUGIN_PATH;
		private static string PACKAGE_PATH;
		private static string PLUGIN_NAME = "VAMMusicLoader";
		private static string EXCEPTION_PREFIX = PLUGIN_NAME + " - Exception caught: ";
		private string lastFileExplorerPath = "Custom";
		private List<string> audioFileExts = new List<string> { ".wav", ".WAV", ".mp3", ".MP3", ".ogg", ".OGG" };
		
		private AudioSource AS;
		private AssetBundleAudioClip mlClip;
		
		// Assetbundle
		public Request BundleRequest = null;
		public bool bundleLoaded = false;
		public bool audioFound = false;
		private string MusicBundleCleanURL = "";
		
		// Storables
		public JSONStorableString MusicBundleURL;

		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "AudioSource")
                {
                    SuperController.LogError("Please add " + PLUGIN_NAME + " on an AudioSource Atom");
                    return;
                }
                
	            PLUGIN_PATH = Utils.GetPluginPath(this);
	            PACKAGE_PATH = Utils.GetPackagePath(this);

	            UIDynamicTextField tmpTextfield;
	            UIDynamicButton tmpButton;
	            
	            Utils.CreateStaticDescriptionText(this, "VAMMLTitle","<color=#000><size=35><b>VAM Music Loader</b></size>\n<size=26>Loads licenced music through asset bundles.</size></color>",false,105,TextAnchor.UpperLeft);

	            MusicBundleURL = new JSONStorableString("MusicBundleURL", "");
	            tmpTextfield = CreateTextField(MusicBundleURL);
	            setupTextField(tmpTextfield, 80f);
	            RegisterString(MusicBundleURL);

	            tmpButton = CreateButton("Select bundle", false);
	            setButtonColor(tmpButton, new Color(0.7f,1f,0.7f,1f) );
	            tmpButton.button.onClick.AddListener(() => { OnSelectMusicBundle(); } );
	            setupButtonWithLayout(tmpButton, 250f);
			
	            tmpButton = CreateButton("Remove bundle", false);
	            setButtonColor(tmpButton, new Color(1f,0.7f,0.7f,1f) );
	            tmpButton.button.onClick.AddListener(() => { OnClearMusicBundle(); } );
	            setupButtonWithoutLayout(tmpButton, 250f, new Vector2(270f, -250f));
	            
	            Utils.CreateStaticDescriptionText(this, "VAMMLTitle","<color=#000><size=32><b>HELP</b></size>\n" +
	                "<size=30>" +
	                "This plugin allows to load a single music track from an assetbundle and use it anywhere in your scene.\n\n" +
	                "Select your assetbundle, then use the sound like you would with an audio source and a trigger.\n\n" +
	                "The sound will be categorized under \"URL / VAMMusicLoader\".\n\n" +
	                "The system allows only one file, so if you are using a bundle with multiple audio files, it will grab the first one found.\n\n" +
	                "Name your assetbundle as usual: <b>something.assetbundle</b>. And ensure you have at least a music file inside. It will search for wav, ogg and mp3. <b>OGG is recommended</b>.\n" +
	                "</size></color>"
					,true,1200,TextAnchor.UpperLeft);
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start()
        {
			try
			{
				CleanAndUpdateMusicBundleUrl(MusicBundleURL.val);
	            
	            // Loading the music bundle if the storable url is set
	            if( MusicBundleURL.val != "" )
	            {
		            BundleRequest = new AssetLoader.AssetBundleFromFileRequest {path = MusicBundleCleanURL, callback = OnBundleLoaded};
		            AssetLoader.QueueLoadAssetBundleFromFile(BundleRequest);
	            }
			}
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
        }
		
		void Update() {

		}

		// **************************
		// Functions
		// **************************
		protected void CleanAndUpdateMusicBundleUrl(string url)
		{
			MusicBundleCleanURL = MusicBundleURL.val;
			if (MusicBundleCleanURL.IndexOf("SELF:/") == 0) MusicBundleCleanURL = MusicBundleCleanURL.Replace("SELF:/", PACKAGE_PATH);
		}

		// **************************
		// Callbacks & similar
		// **************************
		protected void OnBundleLoaded(Request aRequest)
		{
			bundleLoaded = true;
			audioFound = false;

			foreach (string bundleMusicItem in aRequest.assetBundle.GetAllAssetNames())
			{
				string fileName = PathExt.GetFileName(bundleMusicItem);
				string fileExt = PathExt.GetExtension(bundleMusicItem);
				if (audioFileExts.Contains(fileExt))
				{
					mlClip = new AssetBundleAudioClip(aRequest, bundleMusicItem, fileName);
					URLAudioClipManager.singleton.AddClip(mlClip);

					// Getting the first, then stopping. It's a single file system (for now)
					audioFound = true;
					break;
				}
			}

			if (!audioFound)
			{
				SuperController.singleton.Alert("<b><color=orange>"+PLUGIN_NAME+"</color></b>\nNo audio found in your assetbundle.", () => {});
			}
		}
		
		private void OnSelectMusicBundle()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = lastFileExplorerPath;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a music bundle");
			SuperController.singleton.fileBrowserUI.fileFormat = "assetbundle";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			
			List<ShortCut> shortCutList = new List<ShortCut>();
			ShortCut shortAp = new ShortCut();
			shortAp.package = "AddonPackages Filtered";
			shortAp.path = "AddonPackages";
			shortAp.displayName = "Filter: Custom";
			shortAp.packageFilter = "Custom";
			shortCutList.Add(shortAp);
			
			ShortCut shortAps = new ShortCut();
			shortAps.package = "AddonPackages Filtered";
			shortAps.path = "AddonPackages";
			shortAps.displayName = "Filter: Saves";
			shortAps.packageFilter = "Saves";
			shortCutList.Add(shortAps);
			
			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = "Custom\\";
			shortCutCustom.displayName = "Custom";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);
			
			ShortCut shortCutSaves = new ShortCut();
			shortCutSaves.path = "Saves\\";
			shortCutSaves.displayName = "Saves";
			shortCutSaves.flatten = false;
			shortCutList.Add(shortCutSaves);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					UnloadBundle();
						
					lastFileExplorerPath = val.Substring(0, val.LastIndexOfAny(new char[] {'/', '\\'}));
					MusicBundleURL.val = val;
					// I'm forcing the load since the callback will not be triggered if
					// the same path is used ( a creator updating a png and reloading it throught the same URL )
					OnLoadMusicBundleURL(val);
				}
			});
		}

		private void OnClearMusicBundle()
		{
			UnloadBundle();
		}

		private void OnLoadMusicBundleURL( string musicBundleUrl )
		{
			CleanAndUpdateMusicBundleUrl(musicBundleUrl);
			
			BundleRequest = new AssetLoader.AssetBundleFromFileRequest {path = MusicBundleCleanURL, callback = OnBundleLoaded};
			AssetLoader.QueueLoadAssetBundleFromFile(BundleRequest);
		}
		
		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				
				// This allows to ensure that dependencies are properly caught by the package builder
				json["MusicBundleURL"] = SuperController.singleton.NormalizeSavePath(json["MusicBundleURL"].Value);
			}
			return json;
		}
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			UnloadBundle();
		}

		void UnloadBundle()
		{
			if (bundleLoaded)
			{
				if( mlClip != null ) URLAudioClipManager.singleton.RemoveClip(mlClip);

				// Unloading assetbundle
				AssetLoader.DoneWithAssetBundleFromFile(MusicBundleCleanURL);
				
				// Clearing the URL
				MusicBundleURL.valNoCallback = "";
			}
			
			bundleLoaded = false;
		}

		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) disableScrollOnText(target);
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private void setupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}
		
		private void setupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void setButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
    }

    /**** Thanks MacGruber! ****/
    public class AssetBundleAudioClip : URLAudioClip
    {
	    public AssetBundleAudioClip(Request aRequest, string fullPath, string fName)
	    {
		    manager = null;
		    sourceClip = aRequest.assetBundle?.LoadAsset<AudioClip>(fullPath);
		    uid = fName;
		    displayName = fName;
		    category = "VAMMusicLoader";
		    destroyed = false;
		    url = "";
	    }
    }
}